# Overflow :: Play Board Classes
# by Team Soya (see AUTHORS)
#
# Copyright (C) 2005 Arc Riley
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
#

import os
import sys
from math import sqrt
import heapq

from glob import glob 

import soya
import soya.sphere
import widgets

from character import Character, Player, HumanPlayer, AIPlayer
from chargeball import ChargeBall
from poly import Poly

def get_board_path():
  for path in soya.path:
    full_path = os.path.join(path, 'boards')

    if os.path.exists(full_path):
      return full_path

  raise "cant find board files!"

def get_board_files():
  path = get_board_path()

  boards = glob(os.path.join(path, '*.ofb'))

  ret = [] 

  for board in boards:
    path, name = os.path.split(board)
    name, ext = os.path.splitext(name)
    ret.append(name)

  return ret

class Board(soya.World):
  class OverflowCommand:
    def __init__(self, poly, src, prio):
      self.prio = prio
      self.poly = poly
      self.src  = src
    def __cmp__(self, comp):
      return self.prio - comp.prio

  def __init__(self, parent, state, name=None):
    soya.World.__init__(self, parent)

    if 0:
      # XXX for debug
      c = Cube(self)
      c.scale(10, 1, 10)

    self.polygons = []
    self.overflow_heap = []
    self.round_count = 0
    self.reset()
    self.state = state

    if name :
      self.load(name)

  def reset(self) :
    self.title = ''
    self.teams = 0
    self.members = 0
    self.comment = ''

    for poly in self.polygons:
        self.remove(poly)
    
    self.polygons = []
    
    self.starts = []

  def save(self) :
    # check if exists, open prompt to verify overwrite
    info = 'OverFlw1 Board %s %d %d %d %s\n' % (self.name, 
      len(self.polygons), self.teams, self.members, self.comment)
    try :
      fd = open('data/boards/%s.ofb' % self.name, 'w')
      fd.write(info)
    except :
      return
    for poly in self.polygons :
      string = ''
      for point in poly.points :
        string += '%s,%s ' % (str(point[0]), str(point[1]))
      fd.write('%s\n' % string[:-1])
    for team in self.starts :
      fd.write('%s\n' % str(team).replace(',','')[1:-1])
    fd.close()

  def load(self, name) :
    self.reset()
    try :
      fd = open('data/boards/%s.ofb' % name, 'r')
      info = fd.readline()
    except :
      return False
    if info[:15] != 'OverFlw1 Board ' :
      return False 
    #
    # First grab the info
    self.name = name
    self.title, polys, self.teams, self.members, self.comment = \
      info.split(" ",6)[2:7]
    try :
      polys = int(polys)
      self.teams = int(self.teams)
      self.members = int(self.members)
    except ValueError :
      return False
    self.comment = self.comment.replace('\r','').replace('\n','')
    #
    # Next load polygons
    for id in xrange(polys) :
      points = fd.readline().split()
      if len(points)<3 or len(points)>8 :
        return False
      pointlist = []
      for point in points :
        try:
          x, y = point.split(',')
        except ValueError:
          print point
          sys.exit(1)

        try :
          x = float(x)
          y = float(y)
        except ValueError:
          return False
        pointlist.append((x, y))
      self.polygons.append(Poly(self, id, pointlist))
    #
    # Lastly, load the teams
    for team in xrange(self.teams) :
      points = fd.readline().split()
      if len(points) != self.members :
        return False
      pointlist = []
      for point in points :
        if not point.isdigit() :
          return False
        pointlist.append(int(point))          
      self.starts.append(pointlist)
    fd.close()

    #temporary hack for old self.starts format
    #self.starts = [(0.5, 0.5), (0.5, 3.5), (3.5, 0.5), (3.5, 3.5)]

    self.setup()
    return True

  def set_players(self, players):
      self.players = players

  def setup(self):
    # create neighbour graph
    for i in range(len(self.polygons)-1):
      for j in range(i+1, len(self.polygons)):
        p1n = len(self.polygons[i].points)
        p2n = len(self.polygons[j].points)
        for p1 in range(p1n):
          p1x = p1+1
          if p1x == p1n: p1x = 0
          for p2 in range(p2n):
            p2x = p2+1
            if p2x == p2n: p2x = 0
            if ((self.polygons[i].points[p1]  == self.polygons[j].points[p2]) and
               (self.polygons[i].points[p1x] == self.polygons[j].points[p2x])) or \
               ((self.polygons[i].points[p1]  == self.polygons[j].points[p2x]) and
               (self.polygons[i].points[p1x] == self.polygons[j].points[p2])):
                   pol1 = self.polygons[i]
                   pol2 = self.polygons[j]
                   dist = sqrt((pol1.gpoint[0]-pol2.gpoint[0])**2 +
                               (pol2.gpoint[1]-pol2.gpoint[1])**2)
                   pol1.neighbours.append((pol2, dist))
                   pol2.neighbours.append((pol1, dist))

    self.set_picking_context()

    self.calc_center()

  def set_picking_context(self):
    # create a raypicking context of all our polygons 
    # ( dont know why soya needs a center and radius args when passing a list ...)
    self.picking_context = self.RaypickContext(soya.Point(self, 0, 0, 0), 5.0,
                                                items = self.polygons)

  def calc_center(self):
    xsum, zsum = 0, 0

    for poly in self.polygons:
      xsum += poly.gpoint[0]
      zsum += poly.gpoint[1]

    xsum /= len(self.polygons)
    zsum /= len(self.polygons)

    self.gpoint = (xsum, zsum)
    self.center = soya.Point(self, xsum, 0, zsum)

  def raypick_polys(self, pos, vector):
    """ returns picked poly, point, normal """
    result = self.picking_context.raypick(pos, vector)

    if result:
      return (result[0].parent.parent.parent,
              result[0], 
              result[1])

    return False

  def path(self, source_poly, dest_poly, playerid):
      """
      pathfinder - returns a list of ids, including source and destination
      if the path cannot be found, [] is returned
      playerid is the id of the player which requests a path
      """
      #print 'source: ', source_poly.id, 'dest: ', dest_poly.id
      #print 'pathplayerid:',playerid
      ptree = [2000000000] * len(self.polygons)
      bktrace = [-1] * len(self.polygons)
      ptree[source_poly.id] = 0
      visit = [False] * len(self.polygons)

      for ntimes in range(len(ptree)):
        min = 2000000000
        for id, node in enumerate(ptree):
          if node < min and not visit[id]:
            minid = id
            min = node
        visit[minid] = True
        for i in self.polygons[minid].neighbours:
          if ptree[i[0].id] > ptree[minid]+1 and \
                            i[0].playerid in [-1, playerid, -2 - playerid]:
            ptree[i[0].id] = ptree[minid]+1
            bktrace[i[0].id] = minid

      path = []
      parent = dest_poly.id
      while parent != -1:
        path.append(self.polygons[parent])
        parent = bktrace[parent]
      path.reverse()
      if path[0].id != source_poly.id:
        path = []
      return path
 
  def begin_round(self):
    soya.World.begin_round(self)
    self.round_count = self.round_count + 1
    overflows = []
    count = [0] * len(self.polygons)
    while len(self.overflow_heap) > 0 and \
          self.overflow_heap[0].prio == self.round_count:
      overflows.append(heapq.heappop(self.overflow_heap))
    for i in overflows:
      if i.poly.charge_count == len(i.poly.neighbours):
        if i.poly.playerid != -1 and i.poly.playeron and \
                              i.poly.playerid != i.src.playerid:
          plid = i.poly.playerid
          if plid < 0:
              plid = -2 - plid
          self.state.players[plid].throw(i.src)
        i.poly.set_charge(i.src.playerid)
        count[i.poly.id] = 1
      else:
        if count[i.poly.id] == 0:
          if i.poly.playerid != -1 and i.poly.playeron and \
                                i.poly.playerid != i.src.playerid:
            plid = i.poly.playerid
            if plid < 0:
                plid = -2 - plid
            self.state.players[plid].throw(i.src) # OMG, I'm flying!!
          i.poly.set_charge(i.src.playerid)
        else:
          i.poly.set_chargeball()


  def set_overflow(self, poly, src):
    #adds a new overflow to the heap to be processed later
    heapq.heappush(self.overflow_heap, 
                   self.OverflowCommand(poly, src, self.round_count + 33))

  def check_game_over(self):
    self.state.activeplayers -= 1
    if self.state.activeplayers == 1:
      for i in self.state.players:
        if i.game_over == False:
          winner = i.playerid
          break
      widgets.MessagePrompt(text = 'Player %d won' % (winner+1), width = 300,
              height = 200, callback = self.state.game_over)

  # XXX dont think i need this.. (dunk)
  def find_closest_edge(self, point):
    point.convert_to(self)
    px, pz = point.x, point.z
    
    dist = 100000.
    edge = None

    for i,p in enumerate(self.points[:-1]):
      x1, y1, z1 = p
      x2, y2, z2 = points[i + 1]

      x = x1 + ((x2 - x1) / 2)
      z = z1 + ((z2 - z1) / 2)

      dx = px - x
      dz = pz - z
      d = (dx * dx) + (dz * dz)
      if edge is None or d < dist:
        edge = i
        dist = d

    return self.points[edge], self.points[edge + 1]

