from random import random 
from math import sqrt

import soya
from soya.opengl import *

CHARGE_TEXT_MAP = {
  1: (0, 0, 128, 128),
  2: (128, 0, 256, 128),
  3: (256, 0, 384, 128),
  4: (384, 0, 512, 128),
  5: (0, 128, 128, 256),
  6: (128, 128, 256, 256),
  7: (256, 128, 384, 256),
  8: (384, 128, 512, 256),
  9: (0, 256, 128, 384),
  }


# XXX TODO use soya.Sprite instead
class ChargeText(soya.Face):
  """ class for displaying the amount of charges as text
  uses a soya face and changes text coords"""

  def __init__(self, parent):
    soya.Face.__init__(self, parent)

    self.material = soya.Material.get('charge_text')

    sz = 0.5

    self.append(soya.Vertex(self, -sz, 0, -sz))
    self.append(soya.Vertex(self,  sz, 0, -sz))
    self.append(soya.Vertex(self,  sz, 0,  sz))
    self.append(soya.Vertex(self, -sz, 0,  sz))

    self.double_sided = 1

    self.number = -1
    
    self.update()

  def update_tex_coords(self):
    verts = self.vertices

    try:
      tx1, tz1, tx2, tz2 = CHARGE_TEXT_MAP[self.number]
    except KeyError:
      return 

    tx1 /= 512.
    tz1 /= 512.
    tx2 /= 512.
    tz2 /= 512.
    
    verts[0].tex_x = tx1
    verts[0].tex_y = tz1
    verts[1].tex_x = tx2
    verts[1].tex_y = tz1
    verts[2].tex_x = tx2
    verts[2].tex_y = tz2
    verts[3].tex_x = tx1
    verts[3].tex_y = tz2

  def update(self):
    if self.parent.parent.charge_count != self.number:
      self.number = self.parent.parent.charge_count 
      self.update_tex_coords()


class ChargeTextSprite(soya.Sprite):
  def __init__(self, parent):
    soya.Sprite.__init__(self, parent)

    self.number = -1

    self.width = 0.3
    self.height = 0.3
    self.lit = 0 

    self.materials = [
                None, 
                soya.Material.get('n_1'),
                soya.Material.get('n_2'),
                soya.Material.get('n_3'),
                soya.Material.get('n_4'),
                soya.Material.get('n_5'),
                soya.Material.get('n_6'),
                soya.Material.get('n_7'),
                soya.Material.get('n_8'),
                soya.Material.get('n_9'),
                ]
    
    self.update()

  def update(self):
    if self.parent.parent.charge_count != self.number:
      self.number = self.parent.parent.charge_count 
      self.update_material()

  def update_material(self):
    self.material = self.materials[self.number]
    

class ChargeWorldParticles(soya.Smoke):
  """ particle system for a charge """
  
  def __init__(self, parent):
    soya.Smoke.__init__(self, parent)
    self.auto_generate_particle = True
    self.set_sizes((0.1, 0.1), (0.3, 0.3))

  def generate(self, index):
    sx = (random()- 0.5) * .2
    sy = (random())
    sz = (random() - 0.5) * .2
    l = (0.2 * (1.0 + random())) / sqrt(sx * sx + sy * sy + sz * sz) * 0.5
    self.set_particle(index, random()*.5, sx * l, sy * l, sz * l, 0.,0.,0.)


class ChargeWorld(soya.World):
  """ contains all the animation and shapes for displaying a charge """

  def __init__(self, parent):
    soya.World.__init__(self, parent)

    self.solid = 0

    #self.particle_system = ChargeWorldParticles(self)
    self.text = ChargeTextSprite(self)
    self.text.y = 0.5

  def update(self):
    self.text.update()

class CreateChargeParticles(soya.Smoke):
  """ particle system for a charge """
  
  def __init__(self, parent):
    soya.Smoke.__init__(self, parent, removable = 1)
    self.set_sizes((0.1, 0.1), (0.3, 0.3))
    self.material = soya.Material.get('star')
    self.lit = False

  def __generate(self, index):
    sx = (random()- 0.5) * .2
    sy = (random())
    sz = (random() - 0.5) * .2
    l = (0.2 * (1.0 + random())) / sqrt(sx * sx + sy * sy + sz * sz) * 0.5
    self.set_particle(index, random()*.5, sx * l, sy * l, sz * l, 0.,0.,0.)


class BoardChargeball(soya.World):
  def __init__(self, parent):
    soya.World.__init__(self, parent)

    self.chargeball = soya.Sprite(self, material = soya.Material.get('ball'))
    self.chargeball.lit = 0
    self.size = 0.20
    self.chargeball.width = self.size
    self.chargeball.height = self.size

    self.particles = CreateChargeParticles(self)

    self.charge_count = 1

  #def advance_time(self, p):
    #soya.World.advance_time(self, p)

    #if self.size > .40 * self.charge_count:
    #  self.size -= 0.5 * p
    #  self.chargeball.width = self.size
    #  self.chargeball.height = self.size

  def grow(self, charge_count):
    self.particles = CreateChargeParticles(self)
    self.size = float(charge_count)
    self.charge_count = charge_count
    self.chargeball.width = 0.20 + 0.03 * charge_count
    self.chargeball.height = 0.20 + 0.03 * charge_count
    
      

class PolyMesh(soya.PythonCoordSyst):
  """ does the actual drawing of each board piece """

  def __init__(self, parent, faces):
    soya.PythonCoordSyst.__init__(self, parent)

    self.needs_update = True
    self.displaylist = -1

    self.face_data = []

    for face in faces:
      temp_face = []

      for vert in face.vertices:
        temp_face.append((vert.x, vert.y, vert.z))

      self.face_data.append(temp_face)
  
  def batch(self):
    return 1, self, None

  def render(self):
    soya.DEFAULT_MATERIAL.activate()

    glDisable(GL_LIGHTING)
    
    glColor4f(*self.parent.get_color())

    glBegin(GL_TRIANGLES)
    for face in self.face_data:
      for vert in face:
        glVertex3f(*vert)
    glEnd()
   
    points = self.parent.points
   
    glBegin(GL_QUADS)
    for p in range(0, len(points)):
      p1 = points[p]
      if p + 1 >= len(points):
        p2 = points[0]
      else:
        p2 = points[p + 1]

      glVertex3f(p1[0], 0, p1[1])
      glVertex3f(p2[0], 0, p2[1])
      glVertex3f(p2[0], -0.5, p2[1])
      glVertex3f(p1[0], -0.5, p1[1])
    glEnd()
    
    glColor4f(0, 0, 0, 1)
    glLineWidth(4.0)

    glBegin(GL_LINE_LOOP)
    
    for point in points:
      glVertex3f(point[0], 0.001, point[1])

    glEnd()

    glEnable(GL_LIGHTING)


    # TODO displaylists
    # tried it but it only drew one :S
    """
    if self.needs_update:
      if self.displaylist != -1:
        self.displaylist = glGenList()

      glNewList(self.displaylist, GL_COMPILE_AND_EXECUTE)

    
      self.needs_update = False
      glEndList()

    else:
      glCallList(self.displaylist)
    """
      

class Poly(soya.World):
  """ board tile/poly """

  def __repr__(self):
    return "<GamePoly p = %s >" % str(self.points) 

  def __init__(self, parent, id, points = []):
    soya.World.__init__(self, parent)

    self.id = id

    # neighbouring poly's as a list of poly id's
    self.neighbours = []
    # all the outer points in our poly ( all clockwise/anticlockwise ?? ) 
    self.points = points

    # the center of the poly
    xsum = sum([x for x, z in self.points])
    zsum = sum([z for x, z in self.points])
    self.gpoint = (xsum/len(self.points), zsum/len(self.points))
    # the center as a soya.Point
    self.center = soya.Point(self, self.gpoint[0], 0, self.gpoint[1])

    # the owning player
    self.playerid = -1

    # is there a player sitting on it ?
    self.playeron = False
    
    # the amount of charges on the world 
    self.charge_count = 0 

    # the world in which charge gfx exist
    self.charge_world = None

    # charge ball energy
    self.chargeball_val = 0

    # chargeball graphical object
    self.chargeball = None

    # create optimized shapes for each material
    self.create_shape()

  def regenerate(self):
    """ call this when you've changed a point coordinate """
    #self.remove(self.volume)
    self.remove(self.world)
    self.remove(self.mesh)

    xsum = sum([x for x, z in self.points])
    zsum = sum([z for x, z in self.points])
    self.gpoint = (xsum/len(self.points), zsum/len(self.points))
    # the center as a soya.Point
    self.center = soya.Point(self, self.gpoint[0], 0, self.gpoint[1])

    self.create_shape()

  def create_shape(self):
    """ creates an actual soya.Shape for the square """

    # create a temporary world for our shape
    self.world = soya.World(self)
    
    # next we must convert our poly into triangles...

    (x0, z0) = self.points[0]
    v0 = soya.Vertex(self.parent, x0, 0, z0)

    material = soya.Material()
    material.diffuse = (0.5, 0.5, 0.5, 0.0)

    faces = []

    for i in range(len(self.points) - 2):
        (x2, z2) = self.points[i+1]
        (x3, z3) = self.points[i+2]

        v2 = soya.Vertex(self.parent, x2, 0, z2)
        v3 = soya.Vertex(self.parent, x3, 0, z3)

        face = soya.Face(self.world, vertices = [v2, v0, v3])
        face.material = material
        face.double_sided = 1
        faces.append(face)

    # convert the temporary world into a shape object 
    #self.fshape = world.shapify()

    # add the shape to a volume
    #self.volume = soya.Volume(self, self.fshape)

    # this actually does our drawing 
    self.mesh = PolyMesh(self, faces)

  def get_color(self):
    if self.playerid == -1:
      return (0.5, 0.5, 0.5, 0.7)
    elif self.playerid < 0:
      return (0.75, 0.75, 0.75, 0.7)
    else:
      return self.parent.players[self.playerid].color

  def set_charge(self, playerid = -1, value = None, inc = 1):
    """ add a charge to the poly, 
    if value is passed set charge_count to value 
    otherwise increment charge_count by inc"""
    
    if value is not None:
      self.charge_count = value
    else:
      self.charge_count += inc
      if self.charge_count > len(self.neighbours):
        self.overflow()

    self.playerid = playerid

    if self.charge_count > 0:
      if self.charge_world is None:
        self.charge_world = ChargeWorld(self)
        self.charge_world.move(self.center)

      self.charge_world.update()
    
    else:
      if self.charge_world is not None:
        self.remove(self.charge_world)

  def set_chargeball(self):
    if self.playeron:
      player = self.parent.state.players[self.playerid]
      if player.chargeball_val < 9:
        player.chargeball_val += 1
      return
    if self.chargeball_val == 0:
      self.chargeball = BoardChargeball(self)
      self.chargeball.set_xyz(self.gpoint[0], 0.15, self.gpoint[1])
      #self.chargeball = soya.Sprite(self, material = soya.Material.get('ball'))
      #self.chargeball.lit = 0
      #self.chargeball.set_xyz(self.gpoint[0], 0.15, self.gpoint[1])
      #self.chargeball.width = 0.20
      #self.chargeball.height = 0.20
      self.chargeball_val = 1
      print 'CHARGEBALL!!!'
    elif self.chargeball_val < 9:
      self.chargeball_val += 1
      self.chargeball.grow(self.chargeball_val)
      #self.chargeball.width += 0.03
      #self.chargeball.height += 0.03
      #self.chargeball.y += 0.02
      print 'CHARGEBALL!!!'

  def overflow(self):
    self.charge_count = 1
    for i in self.neighbours:
      self.parent.set_overflow(i[0], self)
