from random import random
from math import sin, cos, radians, sqrt

import soya
from soya.cube import Cube
from soya import sdlconst as sdl
from soya import pudding
from soya.opengl import *

import states
import board
import board.poly
import widgets


( SELECT_MODE,
  EDIT_MODE,
  ADD_MODE
  ) = range(3)


class RootWidget(widgets.RootWidget):
  def on_key_down(self, key, mods):
    if key == sdl.K_ESCAPE:
      self.state.try_exit()

    elif key == sdl.K_UP: 
      self.state.move_z_plus()
      return True
    elif key == sdl.K_DOWN:
      self.state.move_z_minus()
      return True
    elif key == sdl.K_LEFT:
      self.state.move_x_plus()
      return True
    elif key == sdl.K_RIGHT:
      self.state.move_x_minus()
      return True
    
    elif key == sdl.K_TAB:
      self.state.toggle_mode()
      return True

    elif key == sdl.K_n:
      self.state.level.reset()
      return True

    elif key == sdl.K_d:
      self.state.remove_polygon()
      return True
    
    elif key == sdl.K_s:
      self.state.level.save()
      return True
    
    elif key == sdl.K_a:
      self.state.add_polygon()
      return True

    return False

  def on_mouse_over(self, x, y, button):
    self.state.mouse_moved(x, y)
    return False

  def on_mouse_down(self, x, y, button):
    self.state.mouse_down(x, y, button)
    return False

  def on_mouse_up(self, x, y, button):
    self.state.mouse_up(x, y, button)  
    return False


class BuildingPolygon(soya.PythonCoordSyst):
  def __init__(self, parent = None):
    soya.PythonCoordSyst.__init__(self, parent)

    self.points = []
    self.last = None

  def add_point(self):
    if len(self.points) and self.last == self.points[0]:
      return True
    else:
      self.points.append(self.last)
      return False

  def set_mouse_pos(self, x, y):
    self.last = (x, y)
    
  def batch(self):
    return 2, self, None

  def render(self):
    glDisable(GL_LIGHTING)
    glDisable(GL_DEPTH_TEST)

    glColor4f(0.0, 1.0, 0.0, 1.0)

    if not len(self.points):
      glBegin(GL_POINTS)
      glVertex3f(self.last[0], 0, self.last[1])
      glEnd()

    else:
      glBegin(GL_LINE_STRIP)

      for point in self.points:
        glVertex3f(point[0], 0, point[1])
        
      if self.last:
        glVertex3f(self.last[0], 0, self.last[1])
          
      glEnd()

    glEnable(GL_LIGHTING)
    glEnable(GL_DEPTH_TEST)


class TileHighlight(soya.PythonCoordSyst):
  def __init__(self, poly, parent = None):
    soya.PythonCoordSyst.__init__(self, parent)

    self.poly = poly
    self.selected_point = None

  def batch(self):
    return 2, self, None


  def select_point(self, point):
    self.selected_point = point

  def render(self):
    glDisable(GL_LIGHTING)
    glDisable(GL_DEPTH_TEST)

    glColor4f(0.0, 1.0, 0.0, 1.0)

    glBegin(GL_POINTS)

    for point in self.poly.points:
      if point == self.selected_point:
        glColor4f(1.0, 0.0, 0.0, 1.0)
        
      glVertex3f(point[0], 0, point[1])

      if point == self.selected_point:
        glColor4f(0.0, 1.0, 0.0, 1.0)


    glEnd()

    glEnable(GL_LIGHTING)
    glEnable(GL_DEPTH_TEST)


class CircleHighlight(soya.PythonCoordSyst):
  def __init__(self, poly, parent = None):
    soya.PythonCoordSyst.__init__(self, parent)
    self.scalef = .5
    
    self.set_xyz(poly.gpoint[0], 0, poly.gpoint[1])

  def batch(self):
    return 2, self, None

  def render(self):
    glDisable(GL_LIGHTING)
    glDisable(GL_DEPTH_TEST)

    glColor4f(1.0, 0.0, 0.0, 0.5)

    glBegin(GL_LINE_LOOP)
    
    for d in range(0, 360, 10):
      x = sin(radians(d)) * self.scalef
      z = cos(radians(d)) * self.scalef

      glVertex3f(x, 0, z)

    glEnd()

    glEnable(GL_LIGHTING)
    glEnable(GL_DEPTH_TEST)


class EditorButton(widgets.PromptButton):
  def __init__(self, **kwargs):
    widgets.PromptButton.__init__(self, height = 40, **kwargs)


class EditorState(states.ApplicationState):
  """ State for when the game is actually running """

  ROOT_WIDGET_CLASS = RootWidget

  def __init__(self, application):
    states.ApplicationState.__init__(self, application)
    
    self.selection = None
    self.highlight = None
    self.mode = SELECT_MODE
    self.building_poly = None
    self.dragging = 0
    

  def setup_scene(self):
    states.ApplicationState.setup_scene(self)

    self.create_background()

    # create widgets
    self.create_widgets()

    self.create_level()

    # make the camera look directly down 
    self.camera.move(self.level.center)
    self.camera.y += 10
    self.camera.look_at(self.level.center)

  def create_background(self):
    """ create background scenery, atmosphere etc... """
    # standard atmosphere 
    atmosphere = soya.SkyAtmosphere()
    atmosphere.ambient = (0.3, 0.3, 0.4, 1.0)
    atmosphere.fog = 1
    atmosphere.fog_type = 0
    atmosphere.fog_start = 40.0
    atmosphere.fog_end = self.camera.back - 10
    atmosphere.fog_color = atmosphere.bg_color = (0.2, 0.5, 0.7, 1.0)
    atmosphere.skyplane = 1
    atmosphere.sky_color = (1.5, 1.0, 0.8, 1.0)

    self.scene.atmosphere = atmosphere
  
  def create_level(self):
    """ create the level object and actually create the game blocks """

    # create the level object 
    self.level = board.Board(self.scene, self)
    self.level.y = 100

  def create_widgets(self):
    """ create game widgets """
    
    toolbox = pudding.container.VerticalContainer(self.root_widget,
                                                  left = 10,
                                                  top = 10,
                                                  height = 100,
                                                  width = 150,
                                                  )

    toolbox.add_child(EditorButton(label = 'Help', func = self.do_help), pudding.EXPAND_HORIZ)
    toolbox.add_child(EditorButton(label = 'New Poly', func = self.do_add_poly), pudding.EXPAND_HORIZ)
    toolbox.add_child(EditorButton(label = 'Open', func = None), pudding.EXPAND_HORIZ)
    toolbox.add_child(EditorButton(label = 'Save', func = self.do_save), pudding.EXPAND_HORIZ)
    toolbox.add_child(EditorButton(label = 'Exit', func = self.try_exit), pudding.EXPAND_HORIZ)

  def do_help(self):
    pass

  def do_add_poly(self):
    self.add_polygon()

  def do_save(self):
    self.level.save()

  def move_z_plus(self): self.camera.z += 1
  def move_z_minus(self): self.camera.z -= 1
  def move_x_plus(self): self.camera.x += 1
  def move_x_minus(self): self.camera.x -= 1

  def remove_polygon(self):
    if self.mode == SELECT_MODE:
      if self.selection:
        self.level.polygons.remove(self.selection)
        self.level.remove(self.selection)
        

  def add_polygon(self):
    if self.mode != ADD_MODE:
      print "switching to add mode"
      self.mode = ADD_MODE
      self.building_poly = BuildingPolygon(self.level)
      self.building_poly.set_mouse_pos(0, 0)

  # mouse events
  def mouse_moved(self, x, y):
    if self.mode == EDIT_MODE:
      if self.dragging:
        self.move_selected_point(x, y)

    elif self.mode == ADD_MODE:
      mouse = self.camera.coord2d_to_3d(x, y)

      nearest = self.get_nearest_point(mouse, self.building_poly)
      if not nearest:
        for poly in self.level.polygons:
          nearest = self.get_nearest_point(mouse, poly)
          if nearest:
            break

      if nearest:
        print "Near Point !!"
        self.building_poly.set_mouse_pos(nearest[0], nearest[1])
      else:
        self.building_poly.set_mouse_pos(mouse.x*10+self.camera.x, -mouse.y*10+self.camera.z)
        
  def mouse_down(self, x, y, button):
    if self.mode == EDIT_MODE:
      self.try_pick(x, y)
      
      if self.highlight.selected_point:
        self.dragging = 1
        self.drag_orig = (x, y)
        self.point_orig = self.highlight.selected_point
        
    elif self.mode == SELECT_MODE:
      self.try_pick(x, y)

    elif self.mode == ADD_MODE:
      self.try_pick(x, y)
      

  def mouse_up(self, x, y, button):
    if self.mode == EDIT_MODE and self.highlight.selected_point:
      self.dragging = 0

  def move_selected_point(self, x, y):
    cur = self.highlight.selected_point

    ind = self.selection.points.index(cur)

    px, py = cur

    dx = (x - self.drag_orig[0]) / 50.0
    dy = (y - self.drag_orig[1]) / 50.0

    nx = self.point_orig[0]+dx
    ny = self.point_orig[1]+dy
    
    self.selection.points[ind] = (nx, ny)
    
    for neighbour in self.selection.neighbours:
      for poly in neighbour:
        if isinstance(poly, board.poly.Poly):
          if cur in poly.points:
            i = poly.points.index(cur)
            poly.points[i] = (nx, ny)
            poly.regenerate()

          
    self.highlight.selected_point = self.selection.points[ind]
    self.selection.regenerate()

  # when we hit <TAB>, switch between modes
  def toggle_mode(self):
    if self.selection:
      self.selection.remove(self.highlight)

      if self.mode == SELECT_MODE:
        print "Editing", self.selection
        self.highlight = TileHighlight(self.selection, self.selection)
        self.mode = EDIT_MODE

      elif self.mode == EDIT_MODE:
        print "Picking", self.selection
        self.highlight = CircleHighlight(self.selection, self.selection)
        self.mode = SELECT_MODE
        
      
  def try_pick(self, x, y):
    # try picking a GamePoly
    mouse = self.camera.coord2d_to_3d(x, y)

    if self.mode == SELECT_MODE:
      self.pick_select_mode(mouse)
    elif self.mode == EDIT_MODE:
      self.pick_edit_mode(mouse)
    elif self.mode == ADD_MODE:
      self.pick_add_mode(mouse)


  def pick_add_mode(self, mouse):
    if self.building_poly.add_point():
      print "Poly done"
      poly = board.poly.Poly(self.level, len(self.level.polygons), self.building_poly.points)
      self.level.polygons.append(poly)
      self.level.remove(self.building_poly)
      self.building_poly = None
      
      self.level.setup()
      
      self.mode = SELECT_MODE


  def get_nearest_point(self, mouse, poly):
    nearest = None
    bestdist = None

    if self.building_poly:
      for point in self.building_poly.points:
        dist = sqrt(
          ((point[0] - self.camera.x - mouse.x*10)*(point[0] - self.camera.x - mouse.x*10))
          +
          ((point[1] - self.camera.z - (-mouse.y)*10)*(point[1] - self.camera.z - (-mouse.y)*10))
          )

        if dist < 0.4:
          if not bestdist or dist < bestdist:
            nearest = point
            bestdist = dist

    for point in poly.points:
      # Thank you Descartes :-)
      dist = sqrt(
        ((point[0] - self.camera.x - mouse.x*10)*(point[0] - self.camera.x - mouse.x*10))
        +
        ((point[1] - self.camera.z - (-mouse.y)*10)*(point[1] - self.camera.z - (-mouse.y)*10))
        )

      if dist < 0.4:
        if not bestdist or dist < bestdist:
          nearest = point
          bestdist = dist

    return nearest
    

  def pick_edit_mode(self, mouse):
    nearest = self.get_nearest_point(mouse, self.selection)
    #print "Nearest nodes :", nearest
    self.highlight.select_point(nearest)

  def pick_select_mode(self, mouse):
    result = self.level.raypick(mouse,
                                self.camera.vector_to(mouse))

    if self.selection and self.highlight:
      self.selection.remove(self.highlight)
      self.highlight = None
    
    if result:
      print result[0].parent.parent,
      #print ' Neighbours: ', str(result[0].parent.parent.parent.neighbours)

      point, normal = result
      poly = point.parent.parent.parent

      self.highlight = CircleHighlight(poly, poly)

      self.selection = poly


    else:
      self.selection = None

  def create_poly(self):
    pass 
      
  def try_exit(self):
    prompt = widgets.YesNoPrompt(text = 'Do you really want to quit?', 
                              callback = self.exit_callback,
                              width = 400,
                              height = 150,
                              )

  def exit_callback(self, prompt):
    if prompt.value:
      self.application.set_state(states.MainMenuState)
       
      
