from random import random, shuffle
from math import sin, cos, radians, fabs

import soya
from soya import sdlconst as sdl
from soya.opengl import *

from soya import pudding 
from soya.pudding.ext.meter import Meter as BaseMeter

import states
import widgets
import board

class PowerMeter(widgets.Meter):
  def __init__(self, parent = None):
    widgets.Meter.__init__(self, parent,
                            left = 10,
                            width = 200,
                            height = 40,
                            min = 0,
                            max = 100)

    self.set_pos_bottom_right(bottom = 10)
                        
  
  def advance_time(self, p):
    #self.value += p * 1.0
    # For TESTING
    self.value += p * 1.0


class ChargeCount(pudding.core.Container):
  def __init__(self, parent, player):
    pudding.core.Container.__init__(self, parent, 
                              left = 10, 
                              width = 50, 
                              height = 50)

    self.set_pos_bottom_right(bottom = 100)

    image = soya.Material.get('ball')

    self.img = pudding.control.Image(self, 
                              left = 0, 
                              top = 0, 
                              bottom = 0,
                              right = 0,
                              material = image)

    self.label = pudding.control.SimpleLabel(self, 
                              label = '')

    self.label.set_pos_bottom_right(0, 0)

    self.player = player
    self.value = -1000

  def begin_round(self):
    pudding.core.Container.begin_round(self)

    if self.value != self.player.chargeball_val:
      self.value = self.player.chargeball_val
      self.label.label = str(self.value)
      self.label.set_pos_bottom_right(0, 0)


class TileChargeCount(pudding.control.SimpleLabel):
  def __init__(self, parent, player):
    pudding.control.SimpleLabel.__init__(self, parent, 
                              left = 40,
                              label= '')

    self.set_pos_bottom_right(bottom = 55)

    self.font = pudding.STYLE.default_font

    self.value = 0

    self.player = player

  def begin_round(self):
    if self.value != self.player.last_poly.charge_count:
      self.value = self.player.last_poly.charge_count
      self.label = str(self.value)
    

class Clock(pudding.control.SimpleLabel):
  def __init__(self, parent = None):
    pudding.control.SimpleLabel.__init__(self, parent, 
                              left = 10, 
                              top = 10,
                              label = "99:99")

    self.font = pudding.STYLE.small_font


class GameCamera(soya.Camera):
  def __init__(self, state):
    self.camera_world = soya.World(state.board)
    self.camera_world.move(state.board.center)
    soya.Camera.__init__(self, self.camera_world)

    self.state = state

    self.angle = -45.
    self.angle_cur = self.angle

    self.lat_angle = 0.
    self.lat_angle_cur = self.lat_angle

    # TODO config!
    self.smooth = True
    self.rotation_speed = 2.

    self.lat_rotation = 0.
    self.rotation = 0.

    self.distance = 10.
    self.back = 200

    self.z = self.distance
    
    self.camera_world.rotate_vertical(self.angle)
    
    self.look_at(self.parent)

  def update(self):
    ang = radians(self.angle)
    y = sin(ang) * self.distance
    z = cos(ang) * self.distance

    self.target_pos.set_xyz(z, y, 0)

    self.target_vec = self.vector_to(self.target_pos)
    self.target_vec.normalize()

  def advance_time(self, p):
    soya.Camera.advance_time(self, p)

    lat_angle_diff = self.lat_angle - self.lat_angle_cur
    
    if fabs(lat_angle_diff) > 1.:
      if self.smooth:
        m = lat_angle_diff * p * .1
        self.camera_world.rotate_lateral(m)
        self.lat_angle_cur += m
      else:
        self.camera_world.rotate_lateral(lat_angle_diff)
        self.lat_angle_cur = self.lat_angle

    angle_diff = self.angle - self.angle_cur

    if fabs(angle_diff) > 1.:
      if self.smooth:
        m = angle_diff * p * .1
        self.camera_world.turn_vertical(m)
        self.angle_cur += m
      else:
        self.camera_world.turn_vertical(angle_diff)
        self.angle_cur = self.angle
     
    dist_diff = self.distance - self.z

    if fabs(dist_diff) > .1:
      m = dist_diff * p * .1
      self.z += m

  def begin_round(self):
    soya.Camera.begin_round(self)
    
    self.lat_angle += self.lat_rotation
    self.angle += self.rotation

    if self.angle < -90:
      self.angle = -90.
    elif self.angle > 0:
      self.angle = 0.

  def start_lat_rotate_left(self):
    self.lat_rotation = -self.rotation_speed

  def start_lat_rotate_right(self):
    self.lat_rotation = self.rotation_speed
  
  def stop_lat_rotate(self):
    self.lat_rotation = 0.
    
  def start_rotate_up(self):
    self.rotation = -self.rotation_speed
    
  def start_rotate_down(self):
    self.rotation = self.rotation_speed

  def stop_rotate(self):
    self.rotation = 0.
    

class BackgroundBuilder:
  def __init__(self, state):
    # do some randomize thing here to choose a function to do the world
    self.scene = state.scene
    self.state = state

    background_funcs = []

    for func in dir(self):
      if not func.startswith('background'):
        continue

      attr = getattr(self, func)

      if not callable(attr):
        continue

      background_funcs.append(attr)

    shuffle(background_funcs)
    background_funcs[0]()

  def background_default(self):
    sun = soya.Light(self.scene)
    sun.directional = 1
    sun.diffuse = (1.0, 0.8, 0.4, 1.0)
    sun.rotate_vertical(-45.0)

    sun.set_xyz(100,100,100)
    
    # standard atmosphere 
    if 0:
      atmosphere = soya.SkyAtmosphere()
      atmosphere.ambient = (0.3, 0.3, 0.4, 1.0)
      atmosphere.fog = 1
      atmosphere.fog_type = 0
      atmosphere.fog_start = 40.0
      atmosphere.fog_end = self.state.camera.back - 10
      atmosphere.fog_color = atmosphere.bg_color = (0.2, 0.5, 0.7, 1.0)
      atmosphere.skyplane = 1
      atmosphere.sky_color = (1.5, 1.0, 0.8, 1.0)
    else:
      atmosphere = soya.SkyAtmosphere()
      
      materials = [
        soya.Material.get('neb_ft'),
        soya.Material.get('neb_rt'),
        soya.Material.get('neb_bk'),
        soya.Material.get('neb_lf'),
        soya.Material.get('neb_dn'),
        soya.Material.get('neb_up')
        ]

      # stops the cracks showing 
      for material in materials:
        material.clamp = 1
     
      atmosphere.set_sky_box(*materials)
   
    self.scene.atmosphere = atmosphere

    """
    material = soya.Material.get('clouds3')

    w = soya.World()
    
    for i in range(0, 50):
      self.random_plane(w, material)

    shape = w.shapify()
    v = soya.Volume(self.state.scene, shape)

    v.move(self.state.board.center)
    """

  def random_sprite(self, parent, material):
    angle = random() * 360.

    distance = 50 + (random() * 150.)

    x = sin(radians(angle)) * distance
    z = cos(radians(angle)) * distance

    face = soya.Sprite(parent, material = material)
    face.set_xyz(x, 0, z)
 
  
  def random_plane(self, parent, material):
    angle = random() * 360.

    distance = 50 + (random() * 150.)

    x = sin(radians(angle)) * distance
    z = cos(radians(angle)) * distance

    face = soya.Face(parent)
    face.append(soya.Vertex(face, -2, -1, 0, 0, 1))
    face.append(soya.Vertex(face, -2,  1, 0, 0, 0))
    face.append(soya.Vertex(face,  2,  1, 0, 1, 0))
    face.append(soya.Vertex(face,  2, -1, 0, 1, 1))
    face.double_sided = 1
    face.material = material
    face.lit = False

    face.x = x
    face.z = z

    face.rotate_lateral(angle)
    
    face.y -= 20 + (random() * 40)

    face.scale(10, 10, 10)


  def _background_really_disgusting_red(self):
    sun = soya.Light(self.scene)
    sun.directional = 1
    sun.diffuse = (2.0, 0.8, 0.4, 1.0)
    sun.rotate_vertical(-45.0)

    sun.set_xyz(100,100,100)
    
    # standard atmosphere 
    atmosphere = soya.SkyAtmosphere()
    atmosphere.ambient = (0.3, 0.3, 0.4, 1.0)
    atmosphere.fog = 1
    atmosphere.fog_type = 0
    atmosphere.fog_start = 40.0
    atmosphere.fog_end = self.state.camera.back - 10
    atmosphere.fog_color = atmosphere.bg_color = (0.7, 0.2, 0.7, 1.0)
    atmosphere.skyplane = 1
    atmosphere.sky_color = (1.5, 1.0, 0.8, 1.0)

    self.scene.atmosphere = atmosphere
 

class RootWidget(widgets.RootWidget):
  def __init__(self, application):
    widgets.RootWidget.__init__(self, application)
  
  def on_key_down(self, key, mods):
    if key == sdl.K_ESCAPE:
      self.state.try_exit()

    elif key == sdl.K_UP:
      self.state.camera.start_rotate_up()
    elif key == sdl.K_DOWN:
      self.state.camera.start_rotate_down()

    elif key == sdl.K_LEFT:
      self.state.camera.start_lat_rotate_left()
    elif key == sdl.K_RIGHT:
      self.state.camera.start_lat_rotate_right()

    elif key == sdl.K_PAGEUP:
      self.state.camera.distance += 1
    elif key == sdl.K_PAGEDOWN:
      self.state.camera.distance -= 1

    elif key == sdl.K_TAB:
      #self.state.playeractive = not self.state.playeractive
      #print 'Active player changed', self.state.playeractive
      pass

  def on_key_up(self, key, mods):
    if key == sdl.K_UP:
      self.state.camera.stop_rotate()
    elif key == sdl.K_DOWN:
      self.state.camera.stop_rotate()

    elif key == sdl.K_LEFT:
      self.state.camera.stop_lat_rotate()
    elif key == sdl.K_RIGHT:
      self.state.camera.stop_lat_rotate()
  
  def on_mouse_down(self, x, y, button):
    if button == 1:
      if self.state.playeractive:
        self.state.try_pick(x, y)
      else:
        self.state.try_pick2(x, y)

    elif button == 2:
      self.state.set_charge(x, y)

    elif button == 3:
      self.state.throw_ball(x, y)


class GameState(states.ApplicationState):
  """ State for when the game is actually running """

  ROOT_WIDGET_CLASS = RootWidget
  MUSIC = 'music/overflow_1.ogg'

  def __init__(self, application, local = False, nrplayers = 1, board_name = ''):
    self.firstpoly = None
    self.playeractive = True
    self.nrplayers = nrplayers
    self.local = local
    self.players = []
    self.board_name = board_name
    states.ApplicationState.__init__(self, application)

  def setup_scene(self):
    # we need more than the base state offers
    #states.ApplicationState.setup_scene(self)

    self.scene = soya.World()

    # create the interesting game objects
    self.create_board()
    self.create_players()

    #self.character1.look_at(self.board)

    self.camera = GameCamera(self)

    self.root_widget = self.ROOT_WIDGET_CLASS(self)
    self.root_widget.add_child(self.camera)
    soya.set_root_widget(self.root_widget)

    self.create_background()

    # create widgets
    self.create_widgets()

  def create_background(self):
    """ create background scenery, atmosphere etc... """
    BackgroundBuilder(self)

  def create_board(self):
    """ create the level object and actually create the game blocks """

    # create the board object 
    self.board = board.Board(self.scene, self, self.board_name)
    self.board.y = 10

  def create_players(self):
    """ create the characters """
    self.players = []

    self.character1 = board.Character(self.board, 1)
    self.character1.y = 0.0
    startpoint = self.board.polygons[self.board.starts[0][0]-1].center
    print '**************', startpoint.x, startpoint.z
    self.character1.set_xyz(startpoint.x, 0.0, startpoint.z)
    self.player1 = board.HumanPlayer(self.character1, self)
    self.player1.position = (startpoint.x, startpoint.z)
    self.player1.sync_char()
    self.player1.set_last_poly()
    self.player1.color = (1.0, 0.0, 0.0, 0.7)
    self.players.append(self.player1)

    self.character2 = board.Character(self.board, 2)
    self.character2.y = 0.0
    startpoint = self.board.polygons[self.board.starts[1][0]-1].center
    self.character2.set_xyz(startpoint.x, 0.0, startpoint.z)
    self.player2 = board.AIPlayer(self.character2, self)
    self.player2.position = (startpoint.x, startpoint.z)
    self.player2.sync_char()
    self.player2.set_last_poly()
    self.player2.color = (0.1, 0.1, 0.9, 0.7)
    self.players.append(self.player2)

    if self.nrplayers == 2:
        self.character3 = board.Character(self.board, 3)
        self.character3.y = 0.0
        startpoint = self.board.polygons[self.board.starts[2][0]-1].center
        self.character3.set_xyz(startpoint.x, 0.0, startpoint.z)
        self.player3 = board.AIPlayer(self.character3, self)
        self.player3.position = (startpoint.x, startpoint.z)
        self.player3.sync_char()
        self.player3.set_last_poly()
        self.player3.color = (0.1, 0.9, 0.1, 0.7)
        self.players.append(self.player3)
        self.character1.look_at(self.character2)
        self.character2.look_at(self.character3)
        self.character3.look_at(self.character1)
    else:
        self.character1.look_at(self.character2)
        self.character2.look_at(self.character1)

    self.board.set_players(self.players)
    self.activeplayers = len(self.players)

    for id, player in enumerate(self.players):
        player.playerid = id
        player.char.playerid = id
        player.last_poly.playerid = -2 - id

  def create_widgets(self):
    """ create game widgets """
    """
      Aug 31 13:45:54 <d1223m> ah yes i want to know game screen widgets
      Aug 31 13:46:06 <tid-wave> Power
      Aug 31 13:46:13 <tid-wave> Chargeball power
      Aug 31 13:46:20 <tid-wave> Number of polygons
      Aug 31 13:46:28 <tid-wave> (owned)
      Aug 31 13:46:37 <tid-wave> TIME
      Aug 31 13:46:47 <antont> (lag)
      Aug 31 13:46:50 <tid-wave> yes
      Aug 31 13:47:02 <tid-wave> and a special place for messages like, OVERFLOW !
      Aug 31 13:47:19 <tid-wave> THROWN AWAY !
      Aug 31 13:47:29 <tid-wave> or such
      Aug 31 13:47:33 <tid-wave> and i guess it's enough
      Aug 31 13:48:28 <d1223m> ok
      Aug 31 13:49:02 <d1223m> tid-wave, nice with the colors :)
      Aug 31 13:49:57 <Arc> power = meter, chargeball = enlarged ball sprite with number next to it
      Aug 31 13:50:41 <Arc> a meter of up to 6 lines showing the number of charges on the board (more important than polygons owned, which varies more quickly) for each team
      Aug 31 13:50:55 <Arc> time countdown for timed games, yea
      Aug 31 13:51:35 <Arc> there's a mechanism in the game mechanic for that
      Aug 31 13:51:39 <Arc> er in the network protocol
      Aug 31 13:52:39 <Arc> the power meter needs to be 80% to drop a charge and needs to refill itself every advance_time (smoother than begin_round)  
    """
          
    self.power_meter = PowerMeter(self.root_widget)
    self.charge_count = ChargeCount(self.root_widget, self.player1)
    self.clock = Clock(self.root_widget)
    self.tilechargecount = TileChargeCount(self.root_widget, self.player1)

    self.logo = widgets.Logo(self.root_widget)
      
  def try_pick(self, x, y):
      # try picking a GamePoly
      mouse = self.camera.coord2d_to_3d(x, y)
      result = self.board.raypick_polys(mouse, 
                                    self.camera.vector_to(mouse))

      if result:
        obj, point, normal = result
        #print obj.id,
        #print ' Neighbours: ', str([i[0].id for i in obj.neighbours])
        
        #if self.firstpoly != None:
        #  print [p.id for p in self.board.path(self.firstpoly, obj,0)]
        
        if self.firstpoly == None:
          self.firstpoly = obj
        
        self.player1.goto_poly(obj)
        print 'PlayerID:', obj.playerid
        print '---------------\n'

  def try_pick2(self, x, y):
      # try picking a GamePoly
      mouse = self.camera.coord2d_to_3d(x, y)
      result = self.board.raypick_polys(mouse, 
                                    self.camera.vector_to(mouse))
      if result:
        obj, point, normal = result

        #print "OBH", obj
        #print obj.id,
        #print ' Neighbours: ', str([i[0].id for i in obj.neighbours])
        
        #if self.firstpoly != None:
        #  print [p.id for p in self.board.path(self.firstpoly, obj,0)]
        
        if self.firstpoly == None:
          self.firstpoly = obj
        
        self.player2.goto_poly(obj)
        print 'PlayerID:', obj.playerid
        print '---------------\n'
      
  def throw_ball(self, x, y):
      # try picking a GamePoly
      mouse = self.camera.coord2d_to_3d(x, y)
      result = self.board.raypick_polys(mouse, 
                                    self.camera.vector_to(mouse))
      if result:
        obj, point, normal = result
        print obj.playeron
        if self.playeractive:
          if self.player1.chargeball_val > 0:
            print self.player1.last_poly.id
            print obj.id
            if self.player1.status == 'suspended' and self.player1.last_poly != obj:
              return
            if self.player1.status == 'suspended':
              obj.playerid = self.player1.playerid
            board.ChargeBall(self.board, self.player1.last_poly, obj,
                                                self.player1.chargeball_val)
            self.player1.chargeball_val = 0
        else:
          if self.player2.chargeball_val > 0:
            if self.player2.status == 'suspended' and self.player2.last_poly != obj:
              return
            if self.player2.status == 'suspended':
              obj.playerid = self.player2.playerid
            board.ChargeBall(self.board, self.player2.last_poly, obj,
                                                self.player2.chargeball_val)
            self.player2.chargeball_val = 0

  def set_charge(self, x, y) :
    if self.playeractive and self.player1.power > 74 :
      self.power_meter.value -= 75.0
    mouse = self.camera.coord2d_to_3d(x, y)
    result = self.board.raypick_polys(mouse, 
                                    self.camera.vector_to(mouse))

    if self.playeractive:
      if self.player1.status == 'idle' and self.player1.power > 74:
        self.player1.last_poly.set_charge(self.player1.playerid)
        self.player1.power -= 75
    else:
      if self.player2.status == 'idle' and self.player2.power > 74:
        self.player2.last_poly.set_charge(self.player2.playerid)
        self.player2.power -= 75
    #if result:
    #  obj = result[0]
    #  for player in self.players:
    #    if player.last_poly == obj and player.status == 'idle':
    #      obj.set_charge(player.playerid)
    #      
    #      break
 
  def try_exit(self):
    prompt = widgets.YesNoPrompt(text = 'Do you really want to quit?', 
                              callback = self.exit_callback,
                              left = 10,
                              top = 10,
                              width = 400,
                              height = 150,
                              )

  def exit_callback(self, prompt):
    if prompt.value:
      self.application.set_state(states.MainMenuState)
 
  def game_over(self, prompt):
    self.application.set_state(states.MainMenuState)

      
