import PIL.Image

import soya
from soya import pudding 
from soya import sdlconst as sdl
from soya.pudding.ext.slicingimage import SlicingImage

import states
import utils
import widgets

class RootWidget(widgets.RootWidget):
  def finish(self):
    self.state.finish()

  def on_mouse_up(self, x, y, button):
    self.finish()

  def on_key_down(self, key, mods):
    if key < 255:
      self.finish()


class IntroImage(SlicingImage):
  FADE_IN = 0
  FADE_OUT = 1
  FADE_NONE = 2

  def __init__(self, parent, image, **kwargs):
    SlicingImage.__init__(self, parent, image, **kwargs)

    self.fade = IntroImage.FADE_NONE

    self.s = 0.0
    self.set_shade((1.0, 1.0, 1.0, self.s))
  
  def set_shade(self, color):
    for slice in self._slices.values():
      slice.shade = color

  def fade_in(self, callback = None):
    self.fade = IntroImage.FADE_IN

    self.callback = callback
  
  def fade_out(self, callback = None):
    self.fade = IntroImage.FADE_OUT

    self.callback = callback

  def advance_time(self, p):
    SlicingImage.advance_time(self, p)

    if self.fade == IntroImage.FADE_IN :
      if self.s < 1.0:
        self.s += p * 0.01

        self.set_shade((1.0, 1.0, 1.0, self.s))
      else:
        self.fade = IntroImage.FADE_NONE
        if self.callback: self.callback()

    elif self.fade == IntroImage.FADE_OUT:
      if self.s > 0.0:
        self.s -= p * 0.01

        self.set_shade((1.0, 1.0, 1.0, self.s))

      else:
        self.fade = IntroImage.FADE_NONE
        if self.callback: self.callback()

  def on_resize(self):
    self.width = self.parent.width
    self.height = self.parent.height

    scale_horiz = self.width / float(self.image_width)
    scale_vert = self.height / float(self.image_height)

    x = 0
    for i, w in enumerate(self._slices_x):
      y = 0
      for j, h in enumerate(self._slices_y):
        im = self._slices[i, j]
        im.left = int(x * scale_horiz)
        im.top = int(y * scale_vert)
        im.width = int((x + w) * scale_horiz) - im.left
        im.height = int((y + h) * scale_vert) - im.top
        im._modified = True
        y += h
      x += w


class IntroState(states.ApplicationState):
  """ State for the beggining of the game, before the main menu """

  ROOT_WIDGET_CLASS = RootWidget
  
  def setup_scene(self):
    states.ApplicationState.setup_scene(self)

    self.root_widget.children.remove(self.logo)
    
    images = [ 
              'splash1.png',
              'splash2.png',
              'splash3.png',
              'splash4.png',
             ]


    self.images = []

    for image in images:
      splash_image = PIL.Image.open(utils.find_image_in_path(image))
      splash_image.convert('RGBA')

      splash = IntroImage(self.root_widget, 
                                splash_image,
                                left = 0, top = 0,
                                right = 0, bottom = 0,
                                )
      splash.anchors = pudding.ANCHOR_ALL

      self.images.append(splash)

    self.last_splash = None
    self.current_image = -1

    self.clock = 0.

    self.finishing = False

    self.do_image()

  def do_image(self):
    self.current_image += 1
    if self.current_image >= len(self.images):
      self.current_image = 0 

    if self.last_splash:
      self.last_splash.fade_out()

    splash = self.images[self.current_image]
    self.last_splash = splash
    splash.fade_in()
 
  def _process_game_events(self):
    for event in self.game_events:
      if event[0] in [sdl.KEYDOWN, sdl.MOUSEBUTTONDOWN]:
        self.application.set_state(states.MainMenuState)

  def begin_round(self):
    states.ApplicationState.begin_round(self)
    self.clock += 1

    if not self.finishing and self.clock > 200:
      self.do_image()
      self.clock = 0 
    
  def finish(self):
    self.finishing = True
    self.last_splash.fade_out(self.really_finish)
  
  def really_finish(self):
    self.application.set_state(states.MainMenuState)
      
