import PIL

import soya
from soya import pudding
from soya import sdlconst as sdl
from soya.opengl import *

from soya.pudding.ext.meter import Meter as BaseMeter
from soya.pudding.ext.slicingimage import SlicingImage

import application
import utils

class ImageButton(pudding.control.Image, pudding.core.InputControl):
  """ clickable image widget
  pass argument "func" as function to call on click 
  """

  def __init__(self, parent = None, func = None, **kwargs):
    pudding.control.Image.__init__(self, parent, **kwargs)
    pudding.core.InputControl.__init__(self)
    self.func = func

    self.shade1 = (0.8, 0.8, 0.8, 0.8)
    self.shade2 = (1.0, 1.0, 1.0, 1.0)

    self.shade = self.shade1

  def process_event(self, event):
    return pudding.core.InputControl.process_event(self, event)

  def on_mouse_up(self, x, y, button):
    if self.func:
      self.func()

  def on_focus(self):
    self.shade = self.shade2
  
  def on_loose_focus(self):
    self.shade = self.shade1

  def on_resize(self):
    self.modified = True
    pudding.control.Image.on_resize(self)


class Button(pudding.control.Button):
  def __init__(self, parent = None, func = None, **kwargs):
    kwargs['height'] = 40
    
    pudding.control.Button.__init__(self, parent, **kwargs)

    self.func = func

    self.label_control.font = pudding.STYLE.small_font
    self.label_control.color = pudding.STYLE.BLUE1

  def on_mouse_up(self, x, y, button):
    if self.func:
      self.func()
    

class Meter(BaseMeter):
  def __init__(self, parent = None, **kwargs):
    BaseMeter.__init__(self, parent, **kwargs)

    self.value = 0

    self.material = soya.Material.get('power_gradient')

  def on_resize(self):
    BaseMeter.on_resize(self)

    self.tex_scale = 1.0 / self.max

  def render_self(self):

    self.material.activate()

    y = 8

    glBegin(GL_QUADS)
    glTexCoord2f(0, 0)
    glVertex2f(0, y)
    glTexCoord2f(self.tex_scale * self.value, 0)
    glVertex2f(int(self.value * self.step), y)
    glTexCoord2f(self.tex_scale * self.value, 1.0)
    glVertex2f(int(self.value * self.step), self.height - y)
    glTexCoord2f(0, 1.0)
    glVertex2f(0, self.height - y)
    glEnd()

    glTranslatef(-8, 0, 0)
    pudding.STYLE.draw_bordered_box(width = self.width + 16, height = self.height)


class RootWidget(pudding.core.RootWidget, pudding.core.InputControl):
  """ root widget with event handling funcs """

  def __init__(self, state):
    self.state = state

    pudding.core.RootWidget.__init__(self, None)
    pudding.core.InputControl.__init__(self)

    self.prompt = None

  def process_event(self, event):
    if self.prompt:
      return self.prompt.process_event(event)

    if not pudding.core.RootWidget.process_event(self, event):
      return pudding.core.InputControl.process_event(self, event)

    return True

  def on_resize(self):
    pudding.core.RootWidget.on_resize(self)


class PromptButton(pudding.control.Button):
  def __init__(self, label = '', func = None, **kwargs):
    self.func = func

    pudding.control.Button.__init__(self, None, label, **kwargs)
    
    self.label_control.font = pudding.STYLE.prompt_font

  def on_click(self):
    if self.func:
      self.func()

  def on_resize(self):
    pudding.control.Button.on_resize(self)
    self.label_control.top -= 3


class PromptConfirmButton(ImageButton):
  def __init__(self, **kwargs):
    material = soya.Material.get('tick')

    ImageButton.__init__(self, material = material, width = 64, height = 64, **kwargs)

    self.shade1 = (0.99, 0.99, 0.99, 0.99)
    self.shade = self.shade1


class PromptDeclineButton(ImageButton):
  def __init__(self, **kwargs):
    material = soya.Material.get('cross')

    ImageButton.__init__(self, material = material, width = 64, height = 64, **kwargs)

    self.shade1 = (0.99, 0.99, 0.99, 0.99)
    self.shade = self.shade1
    
    
class PromptInput(pudding.control.Input):
  def __init__(self, on_return = None, **kwargs):
    pudding.control.Input.__init__(self, **kwargs)

    self.label.font = pudding.STYLE.prompt_font
    self.label.color = pudding.STYLE.GREEN1

    self.on_return_func = on_return 

    self.label.left += 10 
    self.label.right += 10

    self.label.anchors = pudding.ANCHOR_ALL
    self.height = 50

  def on_key_down(self, key, mods):
    if key > 255:
      return 

    pudding.control.Input.on_key_down(self, key, mods)

  def on_return(self):
    if self.on_return_func:
      self.on_return_func()

  def on_resize(self):
    pudding.control.Input.on_resize(self)
    self.label.top += 10
    

class Prompt(pudding.control.Panel, pudding.core.InputControl):  
  """ base class for a prompt. stops all other pudding widgets from 
  capturing events """

  def __init__(self, text = '', buttons = [], callback = None, title = 'Overflow', 
                    **kwargs):

    if application.APP.sound:
      application.APP.play_effect('interface/prompt.wav')

    root_widget = soya.IDLER.root_widget
    root_widget.prompt = self
    
    self.callback = callback

    kwargs['z_index'] = 100
    
    pudding.control.Panel.__init__(self, root_widget, label = title, **kwargs)
    pudding.core.InputControl.__init__(self)

    self.container = pudding.core.Container(self, 
                                            left = 10,
                                            top = 20,
                                            right = 10,
                                            bottom = 10,
                                            anchors = pudding.ANCHOR_ALL,
                                            )
    
    self.message = pudding.control.SimpleLabel(self.container,
                                            label = text,
                                            left = 10, 
                                            top = 10,
                                            right = 10,
                                            bottom = 10,
                                            font = pudding.STYLE.prompt_font,
                                            wrap = True,
                                            )

    self.message.color = pudding.STYLE.GREEN1

    self.create_buttons(buttons)

    self.on_resize()

    self.left = (root_widget.width / 2) - (self.width / 2)
    self.top = (root_widget.height / 2) - (self.height / 2)

    self.animating = True
    self.anim_width = 0
    self.anim_height = 0 

  def render_self(self):
    if self.animating:
      mx = (self.width / 2) - (self.anim_width / 2)
      my = (self.height / 2) - (self.anim_height / 2)

      glTranslatef(mx, my, 0)
      pudding.STYLE.draw_bordered_box(self.anim_width, self.anim_height)
      pudding.STYLE.draw_bordered_box(self.anim_width, self.anim_height)
      pudding.STYLE.draw_bordered_box(self.anim_width, self.anim_height)
      pudding.STYLE.draw_bordered_box(self.anim_width, self.anim_height)

    else:
      # hack to make the box darker
      pudding.STYLE.draw_bordered_box(self.width, self.height)
      pudding.STYLE.draw_bordered_box(self.width, self.height)
      pudding.STYLE.draw_bordered_box(self.width, self.height)
      pudding.STYLE.draw_bordered_box(self.width, self.height)

  def advance_time(self, p):
    if self.animating:
      if int(self.anim_width) != self.width:
        self.anim_width += p * .1 * (self.width - self.anim_width)
      if int(self.anim_height) != self.height:
        self.anim_height += p * .1 * (self.height - self.anim_height)

      if int(round(self.anim_width)) == self.width and int(round(self.anim_height)) == self.height:
        self.animating = False
    else:
      pudding.control.Panel.advance_time(self, p)


  def create_buttons(self, buttons):
    self.button_container = pudding.container.HorizontalContainer(self.container,
                                            left = 10,
                                            height = 30,
                                            right = 10,
                                            )

    self.button_container.set_pos_bottom_right(bottom = 20)
    
    for button in buttons:
      if type(button) is tuple:
        button, style = button
      else:
        style = pudding.EXPAND_BOTH

      self.button_container.add_child(button, style)

  def close(self):
    self.parent.prompt = None
    self.parent.children.remove(self)

    if self.callback:
      self.callback(self)

  def process_event(self, event):
    if self.animating:
      return 
      
    if not pudding.core.InputControl.process_event(self, event):
      return pudding.control.Panel.process_event(self, event)
    
    return True


class MessagePrompt(Prompt):
  def __init__(self, **kwargs):
    buttons = [
          (PromptConfirmButton(func = self.yes), pudding.EXPAND_NONE),
          ]
    
    Prompt.__init__(self, buttons = buttons, **kwargs)
    self.button_container.top -= 20

    self.value = False

  def yes(self):
    self.value = True
    self.close()
  
  def on_key_down(self, key, mods):
    if key == sdl.K_ESCAPE:
      self.no()
      return True
    elif key == sdl.K_RETURN:
      self.yes()
      return True

    return False


class YesNoPrompt(Prompt):
  def __init__(self, **kwargs):
    buttons = [
          (PromptConfirmButton(func = self.yes), pudding.EXPAND_NONE),
          (PromptDeclineButton(func = self.no), pudding.EXPAND_NONE),
          ]
    
    Prompt.__init__(self, buttons = buttons, **kwargs)
    self.button_container.top -= 20

    self.value = False

  def yes(self):
    self.value = True
    self.close()
  
  def no(self):
    self.value = False
    self.close()

  def on_key_down(self, key, mods):
    if key == sdl.K_ESCAPE:
      self.no()
      return True
    elif key == sdl.K_RETURN:
      self.yes()
      return True

    return False

class InputPrompt(Prompt):
  def __init__(self, **kwargs):

    self.input = PromptInput(on_return = self.ok)

    buttons = [
         self.input,
         (PromptButton('OK', self.ok, width = 100), pudding.EXPAND_VERT),
         (PromptButton('Cancel', self.cancel, width = 100), pudding.EXPAND_VERT),
         ]

    Prompt.__init__(self, buttons = buttons, **kwargs)

    self.value = False

  def ok(self):
    self.value = self.input.value
    self.close()

  def cancel(self):
    self.value = False
    self.close()

  def on_key_down(self, key, mods):
    if key == sdl.K_ESCAPE:
      self.cancel()


class MeterPrompt(Prompt):
  def __init__(self, **kwargs):
    Prompt.__init__(self, **kwargs)

  def create_buttons(self, buttons):
    vcont = pudding.container.VerticalContainer(self.container, 
                                            left = 20,
                                            top = self.message.height + 25,
                                            right = 20,
                                            bottom = 15,
                                            )

    self.meter = Meter(height = 40)
    vcont.add_child(self.meter, pudding.EXPAND_BOTH)
    
    vcont.add_child(PromptDeclineButton(func = self.cancel))

  def cancel(self): 
    self.close()


class LoginPrompt(InputPrompt):
  def __init__(self, **kwargs):
    kwargs['width'] = 400
    kwargs['height'] = 245
  
    Prompt.__init__(self, **kwargs)

    self.value = False

  def create_buttons(self, buttons):
    vcont = pudding.container.VerticalContainer(self.container, 
                                            left = 20,
                                            top = self.message.height + 25,
                                            right = 20,
                                            bottom = 15,
                                            )

    self.nick = PromptInput(on_return = self.ok)
    self.key  = PromptInput(on_return = self.ok)

    b = pudding.container.HorizontalContainer()
    b.add_child(pudding.control.SimpleLabel(label = 'Username: ', font = pudding.STYLE.small_font, color = pudding.STYLE.GREEN1), pudding.EXPAND_BOTH)
    b.add_child(self.nick, pudding.EXPAND_HORIZ)

    c = pudding.container.HorizontalContainer()
    c.add_child(pudding.control.SimpleLabel(label = 'Key: ', font = pudding.STYLE.small_font, color = pudding.STYLE.GREEN1), pudding.EXPAND_BOTH)
    c.add_child(self.key, pudding.EXPAND_HORIZ)

    d = pudding.container.HorizontalContainer()
    d.add_child(PromptConfirmButton(func = self.ok))
    d.add_child(PromptDeclineButton(func = self.cancel))
    
    vcont.add_child(b, pudding.EXPAND_BOTH)
    vcont.add_child(c, pudding.EXPAND_BOTH)
    vcont.add_child(d, pudding.EXPAND_BOTH)

  def ok(self):
    if self.nick.value and self.key.value :
      self.value = (self.nick.value, self.key.value)
      self.close()

  def cancel(self):
    self.close()

  def on_key_down(self, key, mods):
    if key == sdl.K_ESCAPE:
      self.cancel()


class ScrollBar(pudding.core.Container, pudding.core.InputControl):
  """ vertical scrollbar widget """

  def _get_value(self):
    return self._value

  def _set_value(self, value):
    self._value = value

  value = property(_get_value, _set_value)
  
  def __init__(self, parent = None, min = 0, max = 100, value = 0, on_change = None, **kwargs):
    pudding.core.Container.__init__(self, parent, **kwargs)
    pudding.core.InputControl.__init__(self)

    self.min = min
    self.max = max
    self._value = value
    self.value_scale = 0.
    self.on_change = on_change

    self.inverted = False

    material = soya.Material.get('scroll1')

    self.up_but = ImageButton(self, 
                                          material = material,
                                          func = self.up_func,
                                          left = 0, 
                                          top = 0,
                                          right = 0, 
                                          height = 20,
                                          )

    self.down_but = ImageButton(self,
                                          material = material,
                                          func = self.down_func,
                                          left = 0,
                                          width = 20,
                                          height = 20,
                                          )

    self.marker = pudding.control.Image(self, 
                                          material = material,
                                          width = 20, 
                                          height = 20,
                                          left = 0,
                                          top = 20)

    self.grab_scroll = False

  def up_func(self, check = True):
    if check and self.inverted:
      self.down_func(False)
      return 

    self.value -= 1
    if self.value < self.min:
      self.value = self.min

    self.on_resize()
    if self.on_change:
      self.on_change()

  def down_func(self, check = True):
    if check and self.inverted:
      self.up_func(False)
      return 
  
    self.value += 1
    if self.value > self.max:
      self.value = self.max

    self.on_resize()
    if self.on_change:
      self.on_change()

  def on_resize(self):
    range = self.max - self.min
    try:
      self.value_scale =  (self.down_but.screen_top - self.up_but.screen_bottom - 20) / float(range)
    except ZeroDivisionError:
      self.value_scale = 1.0 

    if not self.inverted:
      self.marker.top = 20 + (self.value_scale * self.value) 
    else:
      self.marker.set_pos_bottom_right(bottom = 20 + (self.value_scale * self.value))
    
    self.down_but.set_pos_bottom_right(bottom = 0)

  def process_event(self, event):
    if not pudding.core.Container.process_event(self, event):
      return pudding.core.InputControl.process_event(self, event)

    return True

  def on_mouse_down(self, x, y, button):
    self.grab_scroll = True
    return True

  def on_mouse_up(self, x, y, button):
    self.grab_scroll = False
    return True
  
  def on_mouse_over(self, x, y, buttons):
    if self.grab_scroll:
      self.value = int(float(y - 20) / self.value_scale)
      
      print self.value

      self.on_resize()

      if self.on_change:
        self.on_change()

      return True

  def on_loose_focus(self):
    self.grab_scroll = False

    
# TODO rename these more sensibly
  
class ScrollingBufferText(pudding.container.VerticalContainer):
  """ a scrollable text buffer for pudding """
  def __init__(self, parent = None, style = None, **kwargs):
    pudding.container.VerticalContainer.__init__(self, parent, **kwargs)

    w, h = pudding.STYLE.small_font.get_print_size('Wg')
    self.line_size = int(h)

    self.style = style or ScrollingBuffer.FROM_TOP

    self.lines_available = 0

    self.selection = None

    self.lines = [""] 
    self.reset_children = True

    self.scroll_position = 0

    self.update_children()

  def on_resize2(self):
    pudding.container.VerticalContainer.on_resize(self)
  
  def on_resize(self):
    if self.reset_children:
      self.reset_children = False
      self.lines_available = self.height / self.line_size 
      
      self.children = [] 
      
      for i in range(0, self.lines_available):
        line = pudding.control.Label(label = 'test', font = pudding.STYLE.small_font, color = pudding.STYLE.BLUE1) 
        self.add_child(line)

    self.on_resize2()

  def update_children(self):
    for index, child in enumerate(self.children):
      try:
        if self.style == ScrollingBuffer.FROM_TOP:
          x = index
        elif self.style == ScrollingBuffer.FROM_BOTTOM:
          x = -(self.lines_available - index) - self.scroll_position

        child.label = self.lines[x]
      except IndexError:
        child.label = ''


class SelectableScrollingBufferText(ScrollingBufferText):
  class Label(pudding.control.Label):
    def __init__(self, parent = None, func = None, **kwargs):
      pudding.control.Label.__init__(self, parent, **kwargs)
      self.func = func
      self.font = pudding.STYLE.small_font
      self.color = pudding.STYLE.BLUE1

      self.autosize = False

      self.selected = False

    def on_mouse_up(self, x, y, button):
      if self.func:
        self.func(self)

    def render_self(self):
      if self.selected:
        glEnable(GL_BLEND)
        glColor4f(1.0, 0.0, 0.0, 0.2)
        glBegin(GL_QUADS)
        glVertex3f(0, 0, 0)
        glVertex3f(self.width + 5, 0, 0)
        glVertex3f(self.width + 5, self.height + 3, 0)
        glVertex3f(0, self.height + 3, 0)
        glEnd()
        glDisable(GL_BLEND)

      pudding.control.Label.render_self(self)

  def on_resize(self):
    if self.reset_children:
      self.reset_children = False
      self.lines_available = self.height / self.line_size 
      
      self.children = [] 
      
      for i in range(0, self.lines_available):
        line = self.Label(label = 'test', func = self.select_label, height = 20)
        self.add_child(line, pudding.EXPAND_HORIZ)

    self.on_resize2()

  def select_label(self, label):
    if label.label == '':
      return 
  
    if self.selection:
      self.selection.selected = False

    self.selection = label
    self.selection.selected = True
 

class MultiColumnScrollBufferText(ScrollingBufferText):
  class Line(pudding.container.HorizontalContainer, pudding.core.InputControl):
    def __init__(self, parent = None, **kwargs):
      pudding.container.HorizontalContainer.__init__(self, parent, **kwargs)
      pudding.core.InputControl.__init__(self)

      self.padding = 5

      self.selected = False

    def render_self(self):
      if self.selected:
        glEnable(GL_BLEND)
        glColor4f(1.0, 0.0, 0.0, 0.2)
        glBegin(GL_QUADS)
        glVertex3f(0, 0, 0)
        glVertex3f(self.width - 5, 0, 0)
        glVertex3f(self.width - 5, self.height + 3, 0)
        glVertex3f(0, self.height + 3, 0)
        glEnd()
        glDisable(GL_BLEND)

      pudding.container.HorizontalContainer.render_self(self)

    def process_event(self, event):
      pudding.core.InputControl.process_event(self, event)

    def on_mouse_up(self, x, y, button):
      self.parent.set_selection(self)

  class Label(pudding.control.Label):
    def __init__(self, parent = None, **kwargs):
      kwargs['font'] = pudding.STYLE.small_font
      kwargs['color'] = pudding.STYLE.BLUE1
      
      pudding.control.Label.__init__(self, parent, **kwargs)

  def __init__(self, parent = None, columns = 4, **kwargs):
    self.line_controls = []
    ScrollingBufferText.__init__(self, parent, **kwargs)
    
    self.lines = [ [""] * columns ]
    self.columns = columns
    self.padding = 5

    self.selection = None

  def on_resize(self):
    if self.reset_children:
      self.reset_children = False
      self.lines_available = self.height / self.line_size

      self.children = []
      self.line_controls = []

      #print "creating %s lines" % self.lines_available
      
      for i in range(0, self.lines_available):
        line = self.Line(height = self.line_size)

        for j in range(0, self.columns):
          line.add_child(self.Label(label = 'test'), pudding.EXPAND_HORIZ)

        self.add_child(line, pudding.EXPAND_HORIZ)
        self.line_controls.append(line)

    self.on_resize2()

  def update_children(self):
    for index, child in enumerate(self.line_controls):
      for i in range(0, self.columns):
        try:
          child.children[i].label = self.lines[index][i]

        except IndexError:
          child.children[i].label = ''

  def set_selection(self, selline):
    if type(selline) is int:
      try:
        selline = self.line_controls[selline]
      except IndexError:
        selline = None

    if self.selection:
      self.selection.selected = False
    
    self.selection = selline
    self.selection.selected = True
        


class ScrollingBuffer(pudding.container.Container):
  """ a scrollable text buffer for pudding with scrollbars """

  FROM_TOP = 1
  FROM_BOTTOM = 2

  def __init__(self, parent = None, style = None, buffer_class = ScrollingBufferText, **kwargs):
    pudding.container.Container.__init__(self, parent, **kwargs)
    
    self.text = buffer_class(self, style = style)

    self.scroll_bar = ScrollBar(self, width = 20, on_change = self.scroll_change)
    self.scroll_bar.z_index = 5

    self.buffer_len = 256

  def on_resize(self):
    self.scroll_bar.height = self.height
    self.scroll_bar.width = 25
    self.scroll_bar.top = 10
    self.scroll_bar.bottom = 10
    self.scroll_bar.set_pos_bottom_right(right = 5)

    self.text.left = 20
    self.text.top = 20
    self.text.width = self.width - 40
    self.text.height = self.height - 40

    self.text.reset_children = True
    self.text.on_resize()

    self.text.update_children()

  def update_scrollbar(self):
    self.scroll_bar.max = max(0, len(self.text.lines) - self.text.lines_available)
    self.scroll_bar.value = self.text.scroll_position
    self.scroll_bar.on_resize()

  def scroll_change(self):
    self.text.scroll_position = self.scroll_bar.value
    self.text.update_children()
    self.text.on_resize2()
    
  def append(self, line):
    self.text.lines.append(line)
  
    if len(self.text.lines) > self.buffer_len:
      self.text.lines = self.text.lines[:-self.buffer_len]
    
    self.text.update_children()
    self.text.on_resize2()

    self.update_scrollbar()

  def extend(self, lines):
    for line in lines:
      self.text.lines.append(line)

    self.text.update_children()
    self.text.on_resize2()

    self.update_scrollbar()

  def replace(self, lines) :
    self.text.lines = lines
    self.text.update_children()
    self.text.on_resize2()

    self.update_scrollbar()



class Logo(pudding.control.Logo):
  def __init__(self, parent):
    pudding.control.Logo.__init__(self, parent, logo_file = 'logo.png')
    self.z_index = 10

    self.set_pos_bottom_right(bottom = - 70)
    self.anchors = pudding.ANCHOR_BOTTOM_RIGHT

    self.animating = True
    self.alpha = 0.

  def advance_time(self, p):
    pudding.control.Logo.advance_time(self, p)

    if self.animating:
      self.alpha += 0.01 * p

      if self.alpha > 1.:
        self.animating = False

      self.shade = (1.0, 1.0, 1.0, self.alpha)


class Background(SlicingImage):
  def __init__(self, parent):
    image = utils.find_image_in_path('menu_tmp.png')
    SlicingImage.__init__(self, parent, PIL.Image.open(image), left = 0, top =0 , right =0, bottom =0 , anchors = pudding.ANCHOR_ALL)

  def on_resize(self):
    self.width = self.parent.width
    self.height = self.parent.height

    scale_horiz = self.width / float(self.image_width)
    scale_vert = self.height / float(self.image_height)

    x = 0
    for i, w in enumerate(self._slices_x):
      y = 0
      for j, h in enumerate(self._slices_y):
        im = self._slices[i, j]
        im.left = int(x * scale_horiz)
        im.top = int(y * scale_vert)
        im.width = int((x + w) * scale_horiz) - im.left
        im.height = int((y + h) * scale_vert) - im.top
        im._modified = True
        y += h
      x += w


class Background2(pudding.control.Image):
  def __init__(self, parent):
    material = soya.Material.get('background')
  
    pudding.control.Image.__init__(self, parent, 
                                        material = material,
                                        top = 0,
                                        left = 0,
                                        bottom = 0,
                                        right = 0,
                                        )

    pudding.control.anchors = pudding.ANCHOR_ALL
  
  
    self._tex_right = 20
    self._tex_bottom = 20 

  def on_resize(self):
    pudding.control.Image.on_resize(self)
    self._modified = 1
    self.width = soya.root_widget.width
    self.height = soya.root_widget.height
                                        
